from typing import Any, Iterable
from itertools import chain


def pad_indent(x: str, ln: int =0) -> str:
    return x.replace("\n", "\n" + " "*ln)


def form_func_sig(*args: Any, func_name: str="", **kwargs: Any) -> str:
    """
    Forms a function-signature string out of the args and kwargs.

    The returned string has nicely-formatted indentations for things
    like multi-line array reprs.

    Parameter
    ---------
    *args : Any

    func_name : str

    **kwargs : Any

    Returns
    -------
    str

    Examples
    --------
    >>> form_func_sig(2, 3, k=1, z=["apple", None], func_name="func")
    "func(2, 3, k=1, z=['apple', None])"

    >>> x = np.array([[1,2,3], [1 ,2, 3], [1,2,2]])
    >>> form_func_sig(None, x, pizza="mouse", z=x, func_name="func")
    func(None, array([[1, 2, 3],
                      [1, 2, 3],
                      [1, 2, 2]]),
         pizza='mouse', z=array([[1, 2, 3],
                                 [1, 2, 3],
                                 [1, 2, 2]]))

    >>> form_func_sig(2, func_name='f')
    "f(2)"

    >>> form_func_sig(pizza=[[], []], func_name="pizza_shop")
    'pizza_shop(pizza=[[], []])'

    >>> form_func_sig()
    '()'
    """
    assert "\n" not in func_name, "`func_name` cannot contain a newline character"

    def join(x: Iterable) -> str:
        """ Joins the reprs of the elements of `x` using the
        separator ', ' unless the repr contains a newline -
        then the separator begins a newline.
        """
        out = ""
        sep = ""
        for item in x:
            ln_break = "\n" in item
            out += sep
            if ln_break and "\n" not in sep:
                for prev in reversed(out.splitlines()):
                    item = item.replace("\n", "\n" + " "*len(prev))
                    break
            out += item
            sep = ",\n" if ln_break else ", "
        return out

    values = (f"{key}=" + pad_indent(repr(value), len(f"{key}="))
              for key, value in kwargs.items())
    sig = join(chain((repr(i) for i in args), values))
    sig = pad_indent(sig, len(func_name + "("))
    return func_name + "(" + sig + ")"



from bwsi_grader.errors import StudentError
import random

__seed__ = ['cardclass', 'deckclass']

def grade_card(StudentCard):
    from hashlib import sha224
    from bwsi_grader.print import print_passed

    def card_repr(rank, suit):
        _rank_to_str = {11: 'Jack', 12: 'Queen', 13: 'King', 14: 'Ace'}
        _suit_to_str = {'C': 'Clubs', 'H': 'Hearts', 'S': 'Spades', 'D': 'Diamonds'}

        rank = _rank_to_str[rank] if rank in _rank_to_str else str(rank)
        return f'{rank} of {_suit_to_str[suit]}'

    # ensure that we can create a card
    try:
        c = StudentCard(4, 'H')
    except (NameError, TypeError):
        raise AssertionError('Could not create card `Card(4, "H")`')

    # ensure there is a rank and a suit, and that they are correctly set
    try:
        c.rank
    except AttributeError:
        raise AttributeError('Could not access the `rank` attribute after creating `Card(4, "H")`')

    try:
        c.suit
    except AttributeError:
        raise AttributeError('Could not access the `suit` attribute after creating `Card(4, "H")`')

    assert c.rank == 4, f'The `rank` attribute did not match. Created card with `Card(4, "H")`. Expected 4; got {c.rank}'

    assert c.suit == 'H', f'The `suit`attribute did not match. Created card with `Card(4, "H")`. Expected "H"; got {c.suit}'

    assert str(c).lower() == card_repr(4, 'H').lower(), \
        f'__repr__ did not match after being lower-cased; expected {card_repr(4, "H").lower()} got {str(c).lower()}'

    try:
        c = StudentCard(13, 'S')
    except (NameError, TypeError):
        raise AssertionError('Could not create card `Card(13, "S")`')

    assert c.rank == 13, f'The `rank` attribute did not match. Created card with `Card(13, "S")`. Expected 13; got {c.rank}'

    assert c.suit == 'S', f'The `suit`attribute did not match. Created card with `Card(13, "S")`. Expected "S"; got {c.suit}'

    assert str(c).lower() == card_repr(13, "S").lower(), \
        f'__repr__ did not match after being lower-cased; expected {card_repr(13, "S").lower()} got {str(c).lower()}'

    for _ in range(100):
        rank1 = random.randint(2, 14)
        suit = ('C', 'H', 'S', 'D')[random.randint(0, 3)]
        student_card1 = StudentCard(rank1, suit)
        c1 = card_repr(rank1, suit)
        assert str(student_card1).lower() == c1.lower(), \
            f'__repr__ did not match after being lower-cased; expected {c1.lower()} got {str(student_card1).lower()}'

        rank2 = random.randint(2, 14)
        suit = ('C', 'H', 'S', 'D')[random.randint(0, 3)]
        student_card2 = StudentCard(rank2, suit)
        c2 = card_repr(rank2, suit)
        assert str(student_card2).lower() == c2.lower(), \
            f'__repr__ did not match after being lower-cased; expected {c2.lower()} got {str(student_card2).lower()}'

        try:
            assert (student_card1 < student_card2) == (rank1 < rank2)
        except AssertionError:
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            raise AssertionError(f'Testing {c1} < {c2}. Expected {rank1 < rank2}; got {student_card1 < student_card2}')

        try:
            assert (student_card1 > student_card2) == (rank1 > rank2)
        except AssertionError:
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            raise AssertionError(f'Testing {c1} > {c2}. Expected {rank1 > rank2}; got {student_card1 > student_card2}')

        try:
            assert (student_card1 == student_card2) == (rank1 == rank2)
        except AssertionError:
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            raise AssertionError(f'Testing {c1} == {c2}. Expected {rank1 == rank2}; got {student_card1 == student_card2}')

        try:
            assert (student_card1 <= student_card2) == (rank1 <= rank2)
        except AssertionError:
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            raise AssertionError(f'Testing {c1} <= {c2}. Expected {rank1 <= rank2}; got {student_card1 <= student_card2}')

        try:
            assert (student_card1 >= student_card2) == (rank1 >= rank2)
        except AssertionError:
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            # pad
            raise AssertionError(f'Testing {c1} >= {c2}. Expected {rank1 >= rank2}; got {student_card1 >= student_card2}')


    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grade_deck(StudentDeck):
    import random
    from hashlib import sha224
    from bwsi_grader.print import print_passed

    # ensure that we can create a deck
    try:
        deck = StudentDeck()
    except NameError:
        raise AssertionError('Could not create deck by calling `Deck()`. Is `shuffled` optional?')

    # ensure we can create a deck with a shuffled parameter
    try:
        deck = StudentDeck(True)
    except (NameError, TypeError):
        raise AssertionError('Could not create deck with shuffled=True by calling `Deck(True)`')

    # ensure that a deck has a shuffled parameter
    try:
        deck.shuffled
    except AttributeError:
        raise AttributeError('Could not access the `shuffled` parameter of `Deck`')

    # ensure that a shuffle function exists
    try:
        deck.shuffle()
    except AttributeError:
        raise AttributeError('Could not call `shuffle()` on `Deck`')

    # ensure that a deal_card function exists
    try:
        deck.deal_card()
    except AttributeError:
        raise AttributeError('Could not call `deal_card()` on `Deck`')

    # ensure that a reset function exists
    try:
        deck.reset()
    except AttributeError:
        raise AttributeError('Could not call `reset()` on `Deck`')

    # make sure the repr matches
    deck = str(StudentDeck()).lower()
    assert deck == 'deck(dealt 0, shuffled=false)', f'__repr__ did not match; expected "deck(dealt 0, shuffled=False)" got {deck}'

    # make sure the deck deals cards
    deck = StudentDeck()
    try:
        card = deck.deal_card()
        card.rank
        card.suit
    except AttributeError:
        raise AttributeError(f'The Deck should deal Card-instances. Dealt an object of type: {type(card)}'
                             f'\nTo reproduce this error, call:'
                             f'\ndeck = Deck()'
                             f'\ncard = deck.deal_card()'
                             f'\ncard.rank'
                             f'\ncard.suit')

    # make sure the deck has 52 cards
    deck = StudentDeck()
    try:
        for _ in range(52):
            card = deck.deal_card()
            card.rank
            card.suit
    except Exception as e:
        raise StudentError(f'Trying to deal 52 cards produced the following error:'
                           f'\n\t{type(e).__name__}:{e}')

    try:
        # make sure that dealing past the end of the deck returns None
        card = deck.deal_card()
        assert card is None, f'Expected dealing more than 52 cards to return None but returned {card}'
    except Exception as e:
        raise StudentError(f'Expected dealing more than 52 cards to return None.'
                           f'\nTrying to deal 53 cards produced the following error:'
                           f'\n\t{type(e).__name__}:{e}')

    # make sure the counter stayed at 52
    assert str(deck).lower() == 'deck(dealt 52, shuffled=false)', f'Dealing more than 52 cards incremented the counter past 52; got {str(deck).lower()}'

    # make sure the deck has all four suits
    deck.reset()
    suits = set()
    for _ in range(52):
        suits.add(deck.deal_card().suit.upper())
    assert all(suit in suits for suit in ('C', 'H', 'S', 'D')), f'Expected all four suits (C, H, S, D) but found {suits}'

    # make sure that dealing cards updates the counter
    deck = StudentDeck()
    deal_amt = random.randint(1, 52)
    [deck.deal_card() for _ in range(deal_amt)]
    assert str(deck).lower() == f'deck(dealt {deal_amt}, shuffled=false)', f'`deal_card()` did not update the __repr__ correctly; expected "deck(dealt {deal_amt}, shuffled=False)" got {str(deck).lower()}'

    # make sure reset resets the counter
    deck.reset()
    assert str(deck).lower() == f'deck(dealt 0, shuffled=false)', f'`reset()` did not reset the counter correctly; expected "deck(dealt 0, shuffled=false)" from the __repr__; got {str(deck).lower()}'

    # make sure the ranks of all 52 cards in the right order
    deck.reset()
    expected_ranks = list(range(2, 15)) * 4
    actual_ranks = [deck.deal_card().rank for _ in range(52)]
    assert actual_ranks == expected_ranks, f'Expected unshuffled deck to be in the order:\n\t{expected_ranks}\nbut got:\n\t{actual_ranks}'

    # make sure the suits of all 52 cards are correct
    deck.reset()
    suits = [deck.deal_card().suit for _ in range(52)]
    same_as_next = [suits[i] == suits[i+1] for i in range(51)]
    expected = [True]*12 + [False]
    expected *= 4
    expected = expected[:-1]
    assert same_as_next == expected, f'Expected the suits of an unshuffled deck to grouped together but they were not. Suits are:\n\t{suits}'
    assert suits[-2] == suits[-2], f'Expected the suits of an unshuffled deck to grouped together but they were not. Suits are:\n\t{suits}'

    # make sure shuffling does something
    ordered_ranks = list(range(2, 15)) * 4
    ordered_suits = [True]*12 + [False]
    ordered_suits *= 4
    ordered_suits = ordered_suits[:-1]
    deck_in_order = []
    for _ in range(10):
        deck.reset()
        deck.shuffle()
        actual_ranks = [deck.deal_card().rank for _ in range(52)]
        actual_suits = [suits[i] == suits[i+1] for i in range(51)]
        deck_in_order.append(actual_ranks == ordered_ranks and actual_suits == ordered_suits)

    assert not all(deck_in_order), 'Shuffling is expected to change the order of cards but did not'

    # make sure resetting puts the cards back in order
    deck.reset()
    actual_ranks = [deck.deal_card().rank for _ in range(52)]
    actual_suits = [suits[i] == suits[i+1] for i in range(51)]
    assert actual_ranks == ordered_ranks and actual_suits == ordered_suits, '`reset()` should put the deck back in order but it was still shuffled'

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())



r"""
# Pizza Shop
You work at a pizza restaurant, which is starting to accept orders online. You need to
provide a python function that will accept an arbitrary order as its arguments,
and compute the correct price for the order.

Your cost-calculator function should have four arguments:
- pizzas
- drinks
- wings
- coupon


A single pizza order is formed as a list of toppings. For example
- A pizza with no toppings (other than cheese and sauce is: [] (an empty list)
- A pizza with pepperoni and a double order of olives is : ["pepperoni", "olives", "olives"]

*An arbitrary number of pizzas may be ordered, including no pizzas as all*

Drinks come in as a named order (i.e. a keyword argument 'drinks'). If drinks are ordered,
they are specified as a list of sizes (possible sizes: "small", "medium", "large", "tub"). For example, `drinks=["small", "small", "large"]` would indicate an order including two small drinks and a large drink.

Wings come in as a named order as well (keyword argument 'wings'). If wings are ordered,
they are specified as a list of integers (possible sizes: 10, 20, 40, 100). For example, `wings=[20]` would indicate a single order of 20-piece wings.

A coupon may be specified as the keyword argument 'coupon'. It is will be a single
floating point number between 0 and 1. This indicates the fraction of the *pre-tax*
price that is to be subtracted. For example `coupon=.25` would indicate a 25%-off coupon.

A 6.25% tax is applied to every order. The tax is computed on the total cost of the
order *before a coupon is applied*

Round the price to the nearest cent, using the built-in function round. `round(x, 2)` will round `x` to two decimal places.

## Prices
The prices are as follows:

**Pizza**
- \$13.00

**Toppings**
- pepperoni : \$1.00
- mushroom : \$0.50
- olive : \$0.50
- anchovy : \$2.00
- ham : \$1.50

**Drinks**
- small : \$2.00
- medium : \$3.00
- large : \$3.50
- tub : \$3.75

**Wings**
- 10 : \$5.00
- 20 : \$9.00
- 40 : \$17.50
- 100 : \$48.00


## Examples
The following is an order for a plain pizza, a ham and anchovy pizza, two "tub"-sized
drinks, with a 10%-off coupon:
```python
>>> cost_calculator([], ["ham", "anchovy"], drinks=["tub", "tub"], coupon=0.1)
35.61
```

This order consists only of a small drink.
```python
>>> cost_calculator(drinks=["small"])
2.12
```

This is an order of two plain pizzas, a pizza with double-pepperoni, an order of a 10-piece and a 20-piece wings, and a small drink.
```python
>>> cost_calculator([], [], ["pepperoni", "pepperoni"], wings=[10, 20], drinks=["small"])
60.56
```

## Details
You can assume that the front-end of the website will never pass your function erroneous
orders. That is, you will never receive orders for items that do not exist nor
items that contain typos.

Consider defining individual functions responsible for computing
the cost of the pizzas, drinks, and wings, respectively. Have `cost_calculator`
invoke these internally. Alternatively, you can read ahead about dictionaries and make nice
use of these in this problem.

Our `cost_calculator` signature is empty. Part of the assignment is to come up with the
correct function signature!

## Required Concepts
Basic Object Types
- basic math with python

For-Loops and While-Loops
- basic use of for-loops and/or iterables

Basics of Functions
- functions
  - passing a function an arbitrary number of positional arguments
  - specifying default values for arguments
- dictionaries would be useful here
```
"""

from typing import *


__seed__ = ["pizzashop"]


def gen_pizza(toppings, max_num=4):
    from random import randint, choices
    num = randint(0, max_num)
    return [choices(list(toppings), k=randint(0,3)) for i in range(num)]


def gen_items(items, max_num=3):
    from random import randint, choices
    num = randint(0, max_num)
    return choices(list(items), k=num) if num else None


order = List[str], Optional[List[str]], Optional[List[int]], Optional[float]


def grader(student_func: Callable[[*order], float]):
    """ Tests the student's solution to "pizza shope"

    Parameters
    ----------
    student_func: Callable[[order], float]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224
    from collections import OrderedDict

    toppings = dict(pepperoni=1.,
                    mushroom=0.5,
                    olive=0.5,
                    anchovy=2.,
                    ham=1.5)

    dr = dict(small=2.,
              medium=3.,
              large=3.5,
              tub=3.75)

    wi = {10: 5.,
          20: 9,
          40: 17.5,
          100: 48.}

    def soln(*pizzas: List[str],
             drinks: Optional[List[str]] = None,
             wings: Optional[List[int]] = None,
             coupon: Optional[float] = None) -> float:
        # total cost due to pizzas
        cost = 13.*len(pizzas)
        cost += sum(sum(toppings[i] for i in x) for x in pizzas)

        # total cost due to drinks
        if drinks is not None:
            cost += sum(dr[i] for i in drinks)

        # total cost due to wings
        if wings is not None:
            cost += sum(wi[i] for i in wings)

        discount = cost*coupon if coupon is not None else 0.
        cost += cost*.0625
        cost -= discount
        return round(cost, 2)

    std_pizzas = [[[]],
                  [],
                  [["pepperoni"], []],
                  [[],[]]]

    std_orders = [{},
                  OrderedDict([("drinks", ["tub"]), ("coupon", 0.1)]),
                  {"wings": [10, 20]},
                  OrderedDict([("drinks", ["tub", "small"]),
                               ("wings", [10, 10]), ("coupon", 0.1)])]

    for pizzas, items in zip(std_pizzas, std_orders):
        compare_functions(student=student_func, soln=soln,
                          fn_args=tuple(pizzas), fn_kwargs=items)

    for i in range(1000):
        pizzas = gen_pizza(toppings)
        items = {}

        d = gen_items(dr)
        if d is not None:
            items["drinks"] = d

        w = gen_items(wi)
        if w is not None:
            items["wings"] = w

        compare_functions(student=student_func, soln=soln,
                          fn_args=tuple(pizzas), fn_kwargs=items)

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())



"""
# Relaxation Numerical Method
Learn about a numerical approach to solving math problems of the form: f(x) = x.

## Required Concepts
Basic Object Types
- basic math with Python
- the standard library's `math` module
- lists

Basics of Functions
- basic use of functions
- functions are objects

For-Loops and While-Loops
- basic usage of for-loops and while-loops

## Examples
Define the function:

```python
from math import tanh, exp

def f(x):
    return tanh(5*x)


def g(x):
    return 2 - exp(-x)
```

and then calling your relaxation function, passing it this function, an initial guess of $x_{o}=0.5$, and instructing it to perform 5 iterations, should produce the following list:

```python
>>> relaxation_method1(f, xo=-.5, num_it=5)
[-0.5,
 -0.98661429815143031,
 -0.99989620032332682,
 -0.99990910997226823,
 -0.99990912170456125,
 -0.99990912171522284]

 >>> relaxation_method2(func=g, xo=91.0, tol=1e-10, max_it=3)
 [91.0, 2.0, 1.8646647167633872]
```
"""

from typing import *


__seed__ = ["relaxationmethod1", "relaxationmethod2"]


def _reprd_func(f: Callable, string: str) -> Callable:
    class F():
        def __repr__(self): return f"lambda x: {string}"
        def __call__(self, x): return f(x)
        def __eq__(self, x): return True
    return F()


def grader1(student_func: Callable[[Callable[[float], float]], List[float]]):
    """ Tests the student's solution to "Three, Five, Three-Five"

    Parameters
    ----------
    student_func: Callable[[int], Union[str, int]]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from random import randint, random
    from hashlib import sha224
    from math import tanh

    def soln(f, xo, num_it):
        rollout = [xo]
        for i in range(num_it):
            x = f(xo)
            rollout.append(x)
            xo = x
        return rollout

    # pad
    # pad
    # pad
    # pad
    # pad

    print("Finding fixed-points for the function: f(x) = x**2")
    f1 = _reprd_func(lambda x: x**2, "x**2")

    for input_ in range(100):
        input_ = (f1,)
        kwargs = dict(xo=round(random(), 2), num_it=randint(0, 10))
        compare_functions(student=student_func, soln=soln, fn_args=input_, fn_kwargs=kwargs)

    print("Finding fixed-points for the function: f(x) = tanh(4*x)")

    f2 = _reprd_func(lambda x: tanh(4*x), "tanh(4*x)")

    for input_ in range(100):
        input_ = (f2,)
        kwargs = dict(xo=round(random(), 2), num_it=randint(0, 10))
        compare_functions(student=student_func, soln=soln, fn_args=input_, fn_kwargs=kwargs)

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grader2(student_func: Callable[[Callable[[float], float], float, float, int], List[float]]):
    """ Tests the student's solution to "Three, Five, Three-Five"

    Parameters
    ----------
    student_func: Callable[[Callable[[float], float], float, float, int], List[float]]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from random import randint, random
    from hashlib import sha224
    from math import exp

    def error_estimate(x0, x1, x2):
        denom = 2*x1 - x0 - x2
        if denom == 0:
            denom = 1e-14
        return abs((x1 - x2)**2/denom)

    def soln(f, xo, tol, max_it=500):
        rollout = [xo, f(xo), f(f(xo))]
        while max_it > len(rollout) and error_estimate(*rollout[-3:]) > tol:
            rollout.append(f(rollout[-1]))
        return rollout[:max_it]

    # pad
    # pad
    # pad
    # pad
    # pad

    f1 = _reprd_func(lambda x: 2 - exp(-x), "2 - exp(-x)")
    print("Finding fixed-points for the function: f(x) = 2 - exp(-x)")

    for it in [1, 2, 3, 10, 100]:
        input_ = (f1,)
        kwargs = dict(xo=91., tol=1e-10, max_it=it)
        compare_functions(student=student_func, soln=soln, fn_args=input_, fn_kwargs=kwargs)

    for i in range(1000):
        xo = random() * 100
        tol = random()/100
        max_it = randint(1, 1000)
        input_ = (f1,)
        kwargs = dict(xo=xo, tol=tol, max_it=max_it)
        compare_functions(student=student_func, soln=soln, fn_args=input_, fn_kwargs=kwargs)

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())



r"""
# Ciphers
Having fun with cryptographic functions for encrypting messages.

## Required Concepts
Control Flow
- basic usage of for-loops
- if/else statements

Basic Objects
- Built-in functions for strings
- Joining strings
- The modulo function

Sequence Types
- The built-in `index` function
- Indexing sequences
- Checking for membership in a sequence

## Caesar Cipher
A Caesar cipher is a cryptographic function that takes as input a plaintext string and outputs a ciphertext string that is the encoding of the input. Specifically, the Caesar cipher shifts each letter of the alphabet by some fixed amount, wrapping at the end of the alphabet and ignoring whitespace. For example, the string

```
CogWorks is fun
```

encoded using a Caesar cipher with shift 3 becomes

```
FrjZrunv lv ixq
```

Let's break down how this works. We'll order the alphabet so that lowercase letters come before capital letters. This gives us:

```
abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
```

The Caesar cipher will shift each letter by some amount. In our case, we chose to shift by 3 characters. This means we'll map `a` to `d`, `b` to `e`, ..., `w` to `z`, `x` to `A`, ... `W` to `Z`, `X` to `a`, `Y` to `b`, and `Z` to `c`. We can create a mapping just like this:

```
abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
defghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZabc
```

This is essentially what the Caesar cipher is doing. Now if we want to encode our string, we see that `C` maps to `F`, `o` maps to `r`, and so on.

Reversing the process to decode a string is quite straightforward. If we receive the encrypted string

```
Wlys yvjrz
```

and we know that it was encoded using a Caesar cipher with shift 7, we can easily create our reverse substitution alphabet:

```
hijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZabcdefg
abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
```

and decode the message:

```
Perl rocks
```

Fun fact: The ROT13 algorithm is a famous special case of the Caesar cipher. It assumes all letters are the same case. What's special about ROT13 is that it is its own inverse: you apply the exact same function to both encode and decode your string.

#### Breaking the Caesar Cipher
While not very secure, the Caesar cipher can be effective as a very weak form of encryption. Note that to break the encryption, we can simply try shift amounts from 1 through 25 and see which one comes out as correct English. This is a *brute-force* solution.

## Keyword Cipher
A keyword cipher (or substitution cipher) is another cryptographic function. Rather than shift the alphabet by some fixed amount like the Caesar cipher, the Keyword cipher shifts the alphabet based on a keyword or phrase. Let's illustrate with an example.

Suppose our keyphrase is

```
beaverworks is fun
```

and we want to encode the string

```
python rules
```

Like the Caesar cipher, we'll build a substitution alphabet. We'll go through our keyphrase and record the **unique** letters, in order, then append all the rest of the alphabet that wasn't in the keyphrase. For simplicity, we'll only use lowercase letters for this cipher.

Going through the keyphrase we pull out

```
beavrwrksicl
```

Note that each letter is unique, and is in the order we saw it in the keyphrase. The second 'e' we see in 'beaverworks' is simply ignored. Now all we have to do is go through the alphabet and append all the letters that we don't already have. This gives us the following mapping:

```
abcdefghijklmnopqrstuvwxyz
beavrwoksicldfghjmnpqtuxyz
```

Now we can make a straightforward substitution for each letter in our string, `'python rules'`, and get the encoded string

```
hypkgf mqlrn
```

Taking another example, if we receive the encoded string

```
kqbhj khsjhp kpjige
```

and we know that the keyword used to create the cipher was

```
massachusetts institute of technology
```

then we can easily assemble the substitution alphabet

```
abcdefghijklmnopqrstuvwxyz
maschuetinoflgybdjkpqrvwxz
```

and decode the string

```
super secret string
```

#### Breaking the Keyword Cipher
You might be thinking that this is more difficult to decrypt than our Caesar cipher if we don't know the keyword, and you'd be right. With the Caesar cipher, all we had to do to break it was try 25 different shifts and see which one came out in sensible English.

With the keyword cipher, there are `403,291,461,126,605,635,584,000,000` permutations of the alphabet. And this is with only lowercase letters! Even if we could check 2 billion permutations each second, it would take more than 6 billion years to try all the possible permutations here.

Practically, this is where things like frequency analysis come into play, where if we have enough encoded text we can look at the frequency of each letter and try to match these to known frequencies of English. For example, the letter 'e' is very common and if the frequency of 'q' in the encoded text is close to that known frequency, then 'q' probably maps to 'e'. Of course, there are many sophisticated techniques in cryptanalysis worth exploring.
"""

from typing import *


__seed__ = ["cesarcipher", "keywordcipher"]


def grade_cesar_cipher(student_func: Callable[[str, int], str]):
    """ Tests the student's solution to "Cesar Cipher"

    Parameters
    ----------
    student_func:Callable[[str, int], str]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from random import randint, choices
    import string
    from hashlib import sha224

    def encode_caesar_soln(string: str, shift_amt: int) -> str:
        ''' Encodes the specified `string` using a Caesar cipher with shift `shift_amt`

            Parameters
            ----------
            string : str
                The string to encode.

            shift_amt : int
                How much to shift the alphabet by.

            Returns
            -------
            str
                The encoded string.
           '''
        alphabet = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        return ''.join(alphabet[(alphabet.index(i)+shift_amt) % 52]
                       if i in alphabet else i for i in string)
    # pad
    # pad
    # pad
    # pad
    # pad
    # check for some standard cases
    for input_ in (('test', 4), ('CogWorks 2018', 5), ('Perl rocks', 7)):
        compare_functions(student_func, encode_caesar_soln, fn_args=input_)

    for i in range(100):
        test_str = ''.join(choices(string.ascii_letters + string.digits + ' ', k=randint(5, 50)))
        test_shft = randint(0, 52)
        compare_functions(student_func, encode_caesar_soln, fn_args=(test_str, test_shft))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grade_keyword_cipher(student_func: Callable[[str, str], str]):
    """ Tests the student's solution to "Cesar Cipher"

    Parameters
    ----------
    student_func:Callable[[str, str], str]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from random import randint, choices
    import string
    from hashlib import sha224

    def encode_keyword_soln(string, keyword):
        alphabet = 'abcdefghijklmnopqrstuvwxyz'
        string = string.lower()
        keyword = ''.join(keyword.lower().split())
        from collections import OrderedDict
        sub_alphabet = ''.join(k for k in OrderedDict.fromkeys(keyword + alphabet))
        return ''.join(sub_alphabet[alphabet.index(c)] if c in alphabet else c for c in string)

    # pad
    # pad
    # pad
    # pad
    # pad
    # check for some standard cases
    for input_ in (('test', 'test'), ('cogworks 2018', 'python'), ('perl rocks', 'my key')):
        compare_functions(student=student_func, soln=encode_keyword_soln, fn_args=input_)

    for i in range(100):
        test_str = ''.join(choices(string.ascii_lowercase + string.digits + ' ', k=randint(5, 50)))
        kword = ''.join(choices(string.ascii_lowercase + ' ', k=randint(5, 50)))
        compare_functions(student=student_func, soln=encode_keyword_soln, fn_args=(test_str, kword))

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from typing import Callable, Optional, Tuple, Dict, Any, Sequence
from types import FunctionType, MethodType

from . import python
from .python import *

__all__ = ["compare_functions"]


print(f"Using grader version {__version__}")


def compare_functions(student: Callable,
                      soln: Callable,
                      fn_args: Optional[Tuple] = None,
                      fn_kwargs: Optional[Dict[str, Any]] = None,
                      function_name: str ="student_function",
                      comparison_function: Optional[Callable[[Any, Any], bool]] = None):
    """ Compares a student-function with a solution.

    Raises an exception with a descriptive error message that
    describes the nature of the mismatch, or, an error that
    occurred when calling one of the functions.

    Parameters
    ----------
    student : Callable
        The student's function
    soln : Callable
        The solution
    fn_args : Optional[Tuple]
        A tuple of the positional arguments to be unpacked to
        both functions.
    fn_kwargs : Optional[Dict[str, Any]]
        A dictionary of named arguments to be unpacked to both
        functions.
    comparison_function : Optional[Callable[[Any, Any], Union[bool, str]]]
        If specified, this will be used to compare the outputs of
        the respective functions. If it returns `False` and error
        message will print. If it returns a string, that will be printed
        as the error message.

    Raises
    ------
    DeveloperError
        An unexpected error or bad behavior (e.g. argument-mutation)
        occurred when invoking ``soln``.

    StudentError
        The student's result does not match the solution. Or, an
        unexpected error or bad behavior (e.g. argument-mutation)
        occurred when invoking ``fn_kwargs``.

    See Also
    --------
    bwsi_grader.compare.mismatch_error :  Responsible for comparing outputs/

    Examples
    --------
    >>> from bwsi_grader import compare_functions
    >>> def soln(x): return x**2
    >>> def student(x): return x + 2
    >>> compare_functions(student=student, soln=soln, fn_args=(3,))
    StudentError:
    Calling
        student_function(3)
    produced an incorrect result.

    Expected:
    9

    Got:
    5
    """
    from copy import deepcopy
    from .func_sig import form_func_sig, pad_indent
    from .compare import mismatch_error
    from .errors import DeveloperError, StudentError
    from itertools import chain

    if fn_args is None:
        fn_args = tuple()

    if not isinstance(fn_args, tuple):
        raise DeveloperError("`fn_args` must be a tuple "
                             "containing the positional arguments")

    if fn_kwargs is None:
        fn_kwargs = dict()

    if not isinstance(fn_kwargs, dict):
        raise DeveloperError("`fn_kwargs` must be a dictionary "
                             "containing the keyword arguments")

    # make copies of inputs to check for mutations
    args_orig = deepcopy(fn_args)
    kwargs_orig = deepcopy(fn_kwargs)

    sig = form_func_sig(*args_orig, **kwargs_orig, func_name=function_name)

    if not callable(student):
        raise StudentError(f"You must pass the grader your function. Instead, you "
                           f"passed the {type(student)}-type object: {repr(student)}.")

    # run student function, checking for Python errors and mutated inputs
    try:
        student_out = student(*fn_args, **fn_kwargs)
    except Exception as e:
        raise StudentError(f"\nCalling \n\t{pad_indent(sig, ln=4)}\nproduces the following error:"
                           f"\n\t{type(e).__name__}:{e}"
                           f"\nPlease run your function, as displayed above, in your Jupyter "
                           f"notebook to get a detailed stacktrace of the error, and debug "
                           f"your function.")

    # check for mutated inputs
    for f_arg, o_arg in zip(chain(fn_args, fn_kwargs.values()),
                            chain(args_orig, kwargs_orig.values())):
        if mismatch_error(f_arg, o_arg):
            raise StudentError(f"\nCalling \n\t{pad_indent(sig, ln=4)}\nmutated one of the function's "
                               f"input-arguments. The original argument:\n\n{repr(o_arg)}\n\nwas mutated to the "
                               f"value:\n\n{repr(f_arg)}\n\nRevise your function so that you copy this argument "
                               f"before changing it in-place.")

    # run solution function, checking for Python errors and mutated inputs
    try:
        soln_out = soln(*fn_args, **fn_kwargs)
    except Exception as e:
        raise DeveloperError(f"\nCalling \n\t{pad_indent(sig, ln=4)}\nproduces the following error:\n\t{e}\n")

    # check for mutated inputs
    for f_arg, o_arg in zip(chain(fn_args, fn_kwargs.values()),
                            chain(args_orig, kwargs_orig.values())):
        if mismatch_error(f_arg, o_arg):
            raise DeveloperError(f"\nCalling \n\t{pad_indent(sig, ln=4)}\nmutated one of the function's "
                                 f"input-arguments. The original argument:\n\n{repr(o_arg)}\n\nwas mutated to the "
                                 f"value:\n\n{repr(f_arg)}")

    # Return a string that reports any mismatch between function outputs
    # An empty string indicates that the results match
    mismatch_msg = mismatch_error(actual=student_out, desired=soln_out, comparison_function=comparison_function)
    if mismatch_msg:
        raise StudentError(f"\nCalling \n\t{pad_indent(sig, ln=4)}\nproduced an incorrect result.\n\n" + mismatch_msg)


def compare_objects(student_obj: Any, soln_obj: Any,
                    skip_attrs: Optional[Sequence[str]]=None,
                    included_private_attrs: Optional[Sequence[str]] = None):
    """ Compares two object and ensures that they have:
    - Matching names
    - Matching attributes (public only unless otherwise specified)

    Parameters
    ----------
    student_obj : Any,
    soln_obj : Any
    skip_attrs : Optional[Sequential[str]]
    included_private_attrs : Optional[Sequence[str]]

    Raises
    ------
    DeveloperError
        An unexpected error or bad behavior (e.g. non-existent attributes)
        occurred when inspecting ``soln``.

    StudentError
        The student's object does not match the solution.
    """
    from .errors import DeveloperError, StudentError

    def get_name(obj): return obj.__name__ if isinstance(obj, type) else type(obj)

    # ensure soln & student are both instance-objects or both class objects
    if isinstance(soln_obj, type) ^ isinstance(student_obj, type):
        type_ = "class-object" if isinstance(soln_obj, type) else "class-instance"
        other = "class-instance" if isinstance(soln_obj, type) else "class-object"
        raise StudentError(f"You passed the grader a {other}, it expected a {type_}")

    if get_name(student_obj) != get_name(soln_obj):
        raise StudentError(f"You passed the grader an object named {get_name(student_obj)}, expected "
                           f"an object named {get_name(soln_obj)}.")

    name = get_name(soln_obj)

    if skip_attrs is None:
        skip_attrs = []

    if included_private_attrs is None:
        included_private_attrs = []

    if set(included_private_attrs) - set(dir(soln_obj)):
        raise DeveloperError(f"The following attributes are not in the solution-object: "
                             f"{set(included_private_attrs) - set(dir(soln_obj))}")

    for attr in dir(soln_obj):
        if attr in skip_attrs or (attr not in included_private_attrs and attr.startswith("_")):
            continue
        if not hasattr(student_obj, attr):
            obj = getattr(soln_obj, attr)
            type_ = "method" if isinstance(obj, (MethodType, FunctionType)) else "attribute"
            raise StudentError(f"The provided class object, {name}, is missing the {type_}: `{attr}`\n"
                               f"Please revisit your class definition and implement `{name}.{type_}`")


def compare_attributes(*, student_attr: Any, soln_attr: Any, instance_repr: str, attr_name: str):
    from .errors import StudentError
    from .compare import mismatch_error
    mismatch_msg = mismatch_error(actual=student_attr, desired=soln_attr)
    if mismatch_error:
        raise StudentError(f"The attribute `{instance_repr}.{attr_name} returned the wrong value.\n"
                           + mismatch_msg)



"""
# Is it a palindrome?

A palindrome is a word, phrase, or sequence that reads the same backward as forward, e.g., madam or nurses run.

Write code that takes a string and returns `True` if that string is a palindrome.
Your analysis should be case-insensitive and should disregard spaces. E.g. "Race car" should be considered a palindrome,
despite beginning with a capitalized letter and containing a space. None of the strings will contain punctuation.
Thus your function should produce the following behavior:

- "Are we not drawn onward to new era" -> True
- "batman" -> False

## Required Concepts
Basic Object Types
- strings
- checking equality

Sequence Types
- slicing

## Examples
```python
>>> student_func("Race car")
True

>>> student_func("RaCeCaR")
True

>>> student_func("apple")
False
```
"""

from typing import Callable


__seed__ = ["palindrome"]


def grader(student_func: Callable[[str], bool]):
    """ Tests the student's solution to "Is it a Palindrome?"

    Parameters
    ----------
    student_func: Callable[[str], bool]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def soln(x: str) -> bool:
        x = x.lower().replace(" ", "")
        return x == x[::-1]

    # check some standard cases
    for input_ in ("cat", "Race Car", "RaCeCaR", "Python", "nurses run", "HHH"):
        compare_functions(student=student_func, soln=soln, fn_args=(input_,))

    phrases = ['A but tuba',
               'A car a man a maraca',
               'A dog a plan a canal pagoda',
               'A dog A panic in a pagoda',
               'A lad named E Mandala',
               'A man a plan a canal Panama',
               'A man a plan a cat a ham a yak a yam a hat a canalPanama',
               'A new order began a more Roman age bred Rowena',
               'A nut for a jar of tuna',
               'A Santa at Nasa',
               'Desserts I desire not so long no lost one rise distressed',
               'Drat such custard',
               'Evil axis sides reversed is six alive',
               'Gate man sees name garage man sees name tag',
               'Im a fool aloof am I',
               'Socrates Socrates',
               'B',
               'aaabaaa',
               'aaabcaaa']

    for phrase in phrases:
        compare_functions(student=student_func, soln=soln, fn_args=(phrase,))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())



"""# Heat Dispersion
Learn about a numerical approach to simulating the spread of heat through a material

## Required Concepts
For-Loops and While-Loops
- basic usage of for-loops

Accessing Data Along Multiple Dimensions in an Array
- Indexing into, and updating a multi-dimensional array

“Vectorized” Operations: Optimized Computations on NumPy Arrays
- Using vectorized binary functions to perform basic arithmetic on arrays
"""


from typing import *
from numpy import ndarray


__seed__ = ["heatdispersion1", "heatdispersion2"]


def _time_it(student_func: Callable[[ndarray], ndarray],
             soln: Callable[[ndarray], ndarray],
             x: ndarray):
    from timeit import timeit
    from bwsi_grader.errors import StudentError
    soln_time = timeit("soln(x)", number=10, globals=locals())
    student_time = timeit("student_func(x)", number=10, globals=locals())
    if student_time > 3*soln_time:
        msg = """Your function's runtime is too long. Are you using for-loops?
                 
                 Your function is a {}-times slower than the vectorized solution.""".format(round(student_time/soln_time, 1))
        raise StudentError(msg)


def grader(student_func: Callable[[ndarray], ndarray], grade_ver: int):
    """ Tests the student's solution to "Three, Five, Three-Five"

    Parameters
    ----------
    student_func: Callable[[int], Union[str, int]]
       The student's solution function.

    grade_ver: int
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224
    from bwsi_grader.errors import StudentError
    import numpy as np

    if grade_ver not in (1, 2):
        raise StudentError("Invalid grader setting: `grade_ver` was {}, must be 1 or 2".format(grade_ver))

    def soln(u: ndarray) -> ndarray:
        t = np.copy(u)
        t[1:-1, 1:-1] = (u[2:, 1:-1] + u[0:-2, 1:-1] + u[1:-1, 2:] + u[1:-1, 0:-2])/4
        return t
    # pad
    # pad
    # pad
    # pad
    # pad
    # test case from notebook
    u = np.ones((5, 5)) * 100
    u[1:-1, 1:-1] = 0
    compare_functions(student=student_func, soln=soln, fn_args=(u,))
    if grade_ver == 2:
        u = np.ones((100, 100)) * 100
        _time_it(student_func=student_func, soln=soln, x=u)

    # test case no.2
    u = np.ones((3, 5)) * 100
    u[1:-1, 1:-1] = 0
    compare_functions(student=student_func, soln=soln, fn_args=(u,))
    if grade_ver == 2:
        _time_it(student_func=student_func, soln=soln, x=u)

    for i in range(100):
        rand_shape = tuple(np.random.randint(2, 8, 2))
        u = np.round(np.random.rand(*rand_shape), 2)
        compare_functions(student=student_func, soln=soln, fn_args=(u,))

    print_passed(sha224(str.encode(__seed__[grade_ver - 1])).hexdigest())



from bwsi_grader.errors import DeveloperError


def print_passed(hash_str: str):
    """
    Prints the ALL TESTS PASSED message with the included hash.

    The characters 'bw' are prepended to the hash.

    Parameters
    ----------
    hash_str : str
        A 56-character hash to be submitted by students

    Raises
    ------
    DeveloperError
        `hash_str` must be 56 characters.
    """
    if len(hash_str) != 56:
        raise DeveloperError(f"`hash_str` must contain 56 characters. "
                             f"Got {len(hash_str)} characters")
    msg = f"""
============================== ALL TESTS PASSED! ===============================
Your submission code: bw{hash_str}
================================================================================
"""
    print(msg)



"""
# Three, Five, Three-Five
Write code that detects if a number is divisible by 3, 5, both, or neither.
Specifically, write a function that, given `x` (an integer), returns:

    - the string "threefive", if x is a multiple of both 3 and 5.
    - the string "three", if x is a multiple of 3 and not 5.
    - the string "five", if x is a multiple of 5 and not 3.
    - the integer x, if x is not divisible by either 3 or 5.

You may assume that your input is an integer greater than 0.

## Required Concepts
Basic Object Types
- basic math with Python
- strings

Conditional Statements
- basic usage of: if, else, elif


Basics of Functions
- the return statement
- basic function arguments

## Examples
```python
>>> student_func(3)
'three'

>>> student_func(5)
'five'

>>> student_func(15)
'threefive'

>>> student_func(2)
2
```
"""

from typing import Callable, Union


__seed__ = ["threefive"]


def grader(student_func: Callable[[int], Union[str, int]]):
    """ Tests the student's solution to "Three, Five, Three-Five"

    Parameters
    ----------
    student_func: Callable[[int], Union[str, int]]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from random import randint
    from hashlib import sha224

    def soln(x: int) -> Union[str, int]:
        out = ''
        if x % 3 == 0:
            out += 'three'

        if x % 5 == 0:
            out += 'five'
        return out if out else x

    for input_ in (3, 5, 15, 2):
        compare_functions(student=student_func, soln=soln, fn_args=(input_,))

    for i in range(1000):
        compare_functions(student=student_func, soln=soln, fn_args=(randint(1, 1000),))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())



from typing import *
import numpy as np
from numbers import Real


__seed__ = ["polygrad", "gradientdescent", "lossandgradient"]


def grade_polygrad(student_func: Callable[[Tuple[float, ...], float], float]):
    from bwsi_grader.print import print_passed
    from bwsi_grader import compare_functions
    from hashlib import sha224


    def soln(coeffs, x):
        assert all(isinstance(c, Real) for c in coeffs)
        assert isinstance(x, Real)
        coeffs = coeffs[1:]
        return float(sum(coef * ex * (x ** (ex - 1))
                         for coef, ex in zip(coeffs,
                                             range(1, len(coeffs) + 1))))

    # check some standard cases
    for coeffs, x in zip(((), (1.,), (0, 2., 3.5)), (0., 12.4, 5.2)):
        compare_functions(student=student_func, soln=soln, fn_args=(coeffs, x))

    # check random cases

    for i in range(1000):
        num_coeffs = np.random.randint(0, 10)
        coeffs = tuple(float(i) for i in np.round(np.random.rand(num_coeffs)*10, 2))
        x = np.round(np.random.rand(1)*100, 2).item()
        compare_functions(student=student_func, soln=soln, fn_args=(coeffs, x))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grade_gradient_descent(student_func: Callable[[List[float], float, int, float], List[float]]):

    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def soln(poly, step_size=0.1, iterations=10, x=100.):
        """ Returns a list of x-values visited when optimizing `poly` through gradient descent

            Parameters
            ----------
            poly : List[float]
                Polynomial coefficients in increasing order
            step_size : Optional[float], default: 0.1
                The magnitude of the step to take for each update of x
            iterations : Optional[int], default: 10
                After this number of iterations, the grad_descent function should return
            x : Optional[float], default: 100.
                The initial value of x

            Returns
            -------
            x_list : List[float]
                A list of the values of x that were visited, including
                the initial value of x.
        """
        assert isinstance(x, Real)
        assert isinstance(iterations, int)
        assert isinstance(step_size, Real)

        def poly_grad(coeffs, x):
            assert all(isinstance(c, Real) for c in coeffs)
            assert isinstance(x, Real)
            coeffs = coeffs[1:]
            return float(sum(coef * ex * (x ** (ex - 1))
                             for coef, ex in zip(coeffs,
                                                 range(1, len(coeffs) + 1))))
        x_list = [x]
        for _ in range(iterations):
            x -= step_size * poly_grad(poly, x)
            x_list.append(x)
        return x_list

    # test static case
    x = 2.
    coeffs = [1., 2., -4.]
    iterations = 3
    step_size = 0.1

    compare_functions(student=student_func, soln=soln,
                      fn_args=(coeffs,),
                      fn_kwargs=dict(step_size=step_size,
                                     iterations=iterations,
                                     x=x))

    for i in range(1000):
        num_coeffs = np.random.randint(0, 4)
        poly = [float(i) for i in np.random.randint(0, 10, size=(num_coeffs,))]
        x = float((np.random.rand() - 0.5))
        step_size = float(np.random.rand() * 0.01)
        iterations = int(np.random.randint(5, 15))
        _soln = soln(poly, step_size=step_size, iterations=iterations, x=x)

        if not np.isfinite(_soln).all():
            continue # Only test happy paths
        else:
            compare_functions(student=student_func, soln=soln,
                              fn_args=(poly,), fn_kwargs=dict(step_size=step_size,
                                                              iterations=iterations,
                                                              x=x))

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())


def grade_loss_and_gradient(student_func: Callable[[np.ndarray, np.ndarray, np.ndarray], Tuple[float, np.ndarray]]):
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def soln(X, A, y):
        """ Given the input data (X), the collection of model parameters (A), and the
            truth data (y). Compute the loss, and the gradient of the loss (evaluated at A).

            Parameters
            ----------
            X : numpy.ndarray (shape = (N, 4))
                The data for all movies, with the first column of the array
                always equal to 1
            A : np.ndarray (shape = (4,))
                The model parameters: A_bias, A_prod, A_prom, A_book
            y : np.ndarray (shape = (N,))
                The true box office sales for each movie

            Returns
            -------
            out : tuple
                The first entry is the calculated loss and the second entry is the gradient
                evaluated at A.

                The loss should be a single floating-point number
                The gradient should be a numpy array of shape-(4,)

            Fun challenge (optional): For both the loss and the gradient try using matrix
            multiplication and broadcasting to avoid using any for-loops.
        """
        import numpy as np

        assert X.ndim == 2
        assert A.ndim == 1
        assert y.ndim == 1

        N = X.shape[0]
        assert X.shape == (N, 4)
        assert A.shape == (4,)
        assert y.shape == (N,)

        assert np.all(X[:, 0] == 1.)

        y_hat = X.dot(A)
        loss = np.mean((y - y_hat) ** 2)
        grad = -2 * np.mean((y - y_hat) * X.T, axis=1)
        return loss, grad

    X = np.array([[1., 2., 0.1, 8.],
                  [1., 5., 2., 6.]])

    A = np.array([.5, 1., 2., 0.])

    y = np.array((20., 30.))

    compare_functions(student=student_func, soln=soln,
                      fn_args=(X, A, y))

    for _ in range(50):
        p = np.random.randint(1, 10)
        X, A, y = np.random.rand(p, 4).round(2), np.random.rand(4).round(2), np.random.rand(p).round(2)
        X[:, 0] = 1.

        compare_functions(student=student_func, soln=soln,
                          fn_args=(X, A, y))
    print_passed(sha224(str.encode(__seed__[2])).hexdigest())



from numbers import Real
from bwsi_grader.errors import StudentError, DeveloperError

__all__ = ["grade_op", "grade_arithmetic", "grade_backprop", "grade_op_backprop"]
__seed__ = ["op", "arithmetic", "backprop", "opbackprop"]


def _rand_in_range(left, right):
    """Returns random number in [left, right] rounded to
    one decimal."""
    from random import random
    if right <= left:
        raise ValueError("`_rand_in_range` received invalid bounds")
    return round(random() * (right - left) + left, 1)


def try_init(obj, val, creator=None):

    try:
        o = obj(val, creator=creator)
    except Exception as e:
        raise StudentError(f"Calling"
                           f"\n\t`Number({val}, creator={creator})`"
                           f"\nraised the following error:"
                           f"\n"
                           + str(e))
    return o


def grade_op(*, student_Number, student_Add, student_Multiply):
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def check_type(inst, sig):
        if not isinstance(inst, student_Number):
            raise StudentError(
                 f"Calling `{sig}` should return an instance of:"
                 f"\n\t{student_Number}\n"
                 f"Instead, returned type:"
                 f"\n\t{type(z)}")

    def check_creator(inst, type_, sig):
        if not hasattr(inst, "creator"):
            raise StudentError(f"The result of {sig} does not have a creator "
                               f"attribute defined. Every instance of Number should "
                               f"have this attribute."
                               f"\nThe attribute should return "
                               f"`None` in the case that the instance was created by a user and "
                               f"not an operations")
        if isinstance(inst.creator, type):
            raise StudentError(f"The creator of `{sig}` is expected to be an *instance* of:"
                               f"\n\t{type_}\n"
                               f"Instead, found that it is of type:"
                               f"\n\t{type(inst.creator)}"
                               f"\n\n"
                               f"You most likely set the creator to be the operation *class object*. "
                               f"Instead it should be set to the operation *instance* that was used.")
        if not isinstance(inst.creator, type_):
            raise StudentError(f"The creator of `{sig}` is expected to be an instance of:\n\t{type_}\n"
                               f"Instead, found that it is of type:\n\t{type(inst.creator)}")
        if not isinstance(inst, student_Number):
            raise StudentError(
                 f"Calling {sig} should return an instance of {student_Number}, "
                 f"instead returned type: {type(z)}")

    for op, symbol in [(student_Add, "+"), (student_Multiply, "*")]:
        for i in range(20):
            x_ = _rand_in_range(-5, 5)
            y_ = _rand_in_range(-5, 5)

            x = try_init(student_Number, x_)
            y = try_init(student_Number, y_)

            try:
                z = student_Number._op(op, x, y)
            except Exception as e:
                raise StudentError(f"Calling"
                                   f"\n\t{student_Number.__name__}._op({op.__name__}, {x}, {y})"
                                   f"\nproduced the following error:\n\t"
                                   + str(e))

            sig = f"Number({x.data}) {symbol} Number({y.data})"
            check_type(z, sig=sig)
            check_creator(z, op, sig=sig)
            for input_, correct in [(z.creator.a, x), (z.creator.b, y)]:
                if input_ is not correct:
                    raise StudentError(f"Calling"
                                       f"\n\t{student_Number.__name__}._op({op.__name__}, {x}, {y})"
                                       f"\ninappropriately creates a new copy of {correct}.\n"
                                       f"Be sure that you only cast the non-`Number` inputs of `Number._op` as `Number`"
                                       f" instances.")

    for order in ["left", "right"]:
        for op, symbol in [(student_Add, "+"), (student_Multiply, "*")]:
            for i in range(20):
                x_ = _rand_in_range(-5, 5)
                y_ = _rand_in_range(-5, 5)

                if order == "left":
                    x = try_init(student_Number, x_)
                    y = y_
                else:
                    x = x_
                    y = try_init(student_Number, y_)

                try:
                    z = student_Number._op(op, x, y)
                except Exception as e:
                    raise StudentError(f"Calling"
                                       f"\n\t{student_Number.__name__}._op({op.__name__}, {x}, {y})"
                                       f"\n"
                                       f"produced the following error:\n\t"
                                       + str(e))

                sig = f"Number({x_}) {symbol} Number({y_})"
                check_type(z, sig=sig)
                check_creator(z, op, sig=sig)

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grade_arithmetic(*,
                     student_Number,
                     student_Subtract,
                     student_Divide,
                     student_Power):
    import operator as op
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    from bwsi_grader.compare import mismatch_error
    for obj, symb, f, bounds in [(student_Subtract, "-", op.sub, [(-10, 10), (-10, 10)]),
                                 (student_Divide, "/", op.truediv, [(-10, 10), (1, 10)]),
                                 (student_Power, "**", op.pow, [(1, 10), (-10, 10)])]:
        x_bnds, y_bnds = bounds
        for i in range(20):
            x_ = _rand_in_range(*x_bnds)
            y_ = _rand_in_range(*y_bnds)

            x = try_init(student_Number, x_)
            y = try_init(student_Number, y_)

            try:
                correct = f(x_, y_)
            except Exception:
                raise DeveloperError("Invalid operation: {f}({x}, {y})")

            try:
                inst = obj()
                out = inst.__call__(x, y)
            except Exception as e:
                raise StudentError(
                    f"Calling"
                    f"\n\t{obj.__name__}().__call__({x}, {y})"
                    f"\nraised the following error:"
                    f"\n\t{e}")

            mismatch = mismatch_error(actual=out, desired=correct)
            if mismatch:
                raise StudentError(f"Calling"
                                   f"\n\t{obj.__name__}().__call__({x}, {y})\n"
                                   f"produced the wrong result. {mismatch}")

            for correct, attr_name in [(x, "a"), (y, "b")]:
                if not hasattr(inst, attr_name):
                    raise StudentError(f"Calling"
                                       f"\n\top = {obj.__name__}()"
                                       f"\n\top.__call__({x}, {y})\n" 
                                       f"failed to bind {correct} to `op.{attr_name}`")
                attr = getattr(inst, attr_name)
                if not isinstance(attr, student_Number):
                    raise StudentError(f"Calling"
                                       f"\n\top = {obj.__name__}()"
                                       f"\n\top.__call__({x}, {y})\n" 
                                       f"set `{attr}` to `op.{attr_name}`. Instead, it should have "
                                       f"set `{correct}` (type: Number).")
                if attr is not correct:
                    raise StudentError(f"Calling"
                                       f"\n\top = {obj.__name__}()"
                                       f"\n\top.__call__({x}, {y})\n" 
                                       f"\ninappropriately creates a new copy of {correct}.\n"
                                       f"op.{attr_name} should return the same instance of {correct}")
    print_passed(sha224(str.encode(__seed__[1])).hexdigest())


def grade_backprop(*, student_Number, student_Add, student_Multiply):
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def _run(gen): return [i for i in gen]

    def grad_check(inst, expected, seq_steps, name):
        if (inst.grad is not None and not isinstance(inst.grad, (int, float))) or \
           (inst.grad != expected) if expected is not None else (inst.grad is not expected):
            seq_steps = "\n\t".join(seq_steps)
            raise StudentError(
                f"After calling:"
                f"\n\t{seq_steps}"
                f"\n`{name}.grad` should return:"
                f"\n\t{expected}"
                f"\nInstead, it returned:"
                f"\n\t{inst.grad}")

    def mutation_check(inst, expected, seq_steps, name):
        if inst.data != expected:
            seq_steps = "\n\t".join(seq_steps)
            raise StudentError(
                f"Calling:"
                f"\n\t{seq_steps}"
                f"\nmutated the value stored in `{name}`."
                f"\n`{name}` stored {expected}; now it stores {inst.data}")

    def call_check(func, inst, seq_steps):
        try:
            return func(inst)
        except Exception as e:
            seq_steps = "\n\t".join(seq_steps)
            raise StudentError(
                f"Calling:"
                f"\n\t{seq_steps}"
                f"\nraised the following error:"
                f"\n\t{e}")

    def value_check(inst, expected, seq_steps, name):
        from math import isclose
        if not isclose(inst.data, expected, abs_tol=1e-5, rel_tol=1e-5):
            seq_steps = "\n\t".join(seq_steps)
            raise StudentError(
                f"After calling:"
                f"\n\t{seq_steps}"
                f"\n`{name}` should return:"
                f"\n\tNumber({expected})"
                f"\nInstead, it returned:"
                f"\n\tNumber({inst.data})")
    from random import choice

    """
    Checks:
    x = Number(_x)
    x.backprop(g)
    x.grad == g
    """
    for i in range(100):
        _x = _rand_in_range(-10, 10) if i > 0 else 2.
        grad = choice([_rand_in_range(-10, 10), None]) if i > 0 else None
        _g = grad if grad is not None else 1.

        x = try_init(student_Number, _x)
        seq_steps = [f"x = Number({_x})"]
        grad_check(x, expected=None, seq_steps=seq_steps, name="x")

        seq_steps.append(f"x.backprop({grad})" if grad is not None else f"x.backprop()")
        call_check(lambda inst: inst.backprop(_g), inst=x, seq_steps=seq_steps)
        mutation_check(x, expected=_x, seq_steps=seq_steps, name="x")
        grad_check(x, expected=_g, seq_steps=seq_steps, name="x")

        seq_steps.append("x.null_gradients()")
        call_check(lambda inst: inst.null_gradients(), inst=x, seq_steps=seq_steps)
        mutation_check(x, expected=_x, seq_steps=seq_steps, name="x")
        grad_check(x, expected=None, seq_steps=seq_steps, name="x")

    """
    Checks:
    y = Number(_y)
    x = 2 + y
    out = x * x
    out.backprop(g)
    
    out.grad == g
    x.grad == 2*g*x
    y.grad = x.grad
    """
    for i in range(100):
        _y = _rand_in_range(-10, 10) if i > 0 else 3.
        grad = choice([_rand_in_range(-10, 10), None]) if i > 0 else None
        _g = grad if grad is not None else 1.

        y = try_init(student_Number, _y)
        _x = 2 + _y

        _g = grad if grad is not None else 1.

        seq_steps = [f"y = Number({_y})"]
        grad_check(x, expected=None, seq_steps=seq_steps, name="x")
        grad_check(y, expected=None, seq_steps=seq_steps, name="y")

        seq_steps.append(f"x = 2 + y")
        x = call_check(lambda inst: student_Number._op(student_Add, 2, inst), inst=y, seq_steps=seq_steps)
        value_check(x, expected=_x, seq_steps=seq_steps, name="x")

        seq_steps.append(f"out = x * x")
        _out = _x * _x
        out = call_check(lambda inst: student_Number._op(student_Multiply, inst, inst), inst=x, seq_steps=seq_steps)

        vars_ = [y, x, out]
        vals = [_y, _x, _out]
        names = ["y", "x", "out"]

        _run(mutation_check(var, expected=val, seq_steps=seq_steps, name=name)
             for var, val, name in zip(vars_, vals, names))

        value_check(out, expected=_out, seq_steps=seq_steps, name="out")

        seq_steps.append(f"out.backprop({grad})" if grad is not None else f"out.backprop()")
        call_check(func=lambda inst: inst.backprop(_g), inst=out, seq_steps=seq_steps)
        _run(mutation_check(var, expected=val, seq_steps=seq_steps, name=name)
             for var, val, name in zip(vars_, vals, names))

        grad_check(out, expected=_g, seq_steps=seq_steps, name="out")
        grad_check(x, expected=2*_x*out.grad, seq_steps=seq_steps, name="x")
        grad_check(y, expected=x.grad, seq_steps=seq_steps, name="y")

        seq_steps.append("out.null_gradients()")
        call_check(lambda inst: inst.null_gradients(), inst=out, seq_steps=seq_steps)

        _run(mutation_check(var, expected=val, seq_steps=seq_steps, name=name)
             for var, val, name in zip(vars_, vals, names))

        _run(grad_check(var, expected=None, seq_steps=seq_steps, name=name)
             for var, val, name in zip(vars_, vals, names))

    print_passed(sha224(str.encode(__seed__[2])).hexdigest())


def grade_op_backprop(*, student_Number, student_Subtract, student_Divide, student_Power):
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    import numpy as np
    from math import isclose

    def sub_a(a, b): return 1.

    def sub_b(a, b): return -1.

    def div_a(a, b): return 1. / b

    def div_b(a, b): return -1 * a / (b ** 2)
    
    def pow_a(a, b): return b * (a ** (b - 1))

    def pow_b(a, b): return (a ** b) * np.log(a)

    all_ops = [student_Subtract, student_Divide, student_Power]
    partials = [(sub_a, sub_b), (div_a, div_b), (pow_a, pow_b)]
    bounds = [((-10, 10), (-10, 10)),
              ((-3, 3), (1, 3)),
              ((1, 3), (-2, 2))]
    for Op, partial, (a_bnds, b_bnds) in zip(all_ops, partials, bounds):
        for i in range(100):
            x = student_Number(_rand_in_range(*a_bnds) if i > 0 else 1.)
            y = student_Number(_rand_in_range(*b_bnds) if i > 0 else 3.)

            op = Op()
            op(x, y)
            for n, letter in enumerate(["a", "b"]):
                try:
                    actual = op.partial_a() if n == 0 else op.partial_b()
                except Exception as e:
                    raise StudentError(
                        f"After calling:"
                        f"\n\t>>> f = {Op.__name__}()"
                        f"\n\t>>> f({x}, {y})"
                        f"\n Calling:"
                        f"\n`{'f.partial_a()' if n == 0 else 'f.partial_b()'}` raised the error:"
                        f"\n\t{type(e).__name__}: {e}")
                try:
                    expected = partial[n](x.data, y.data)
                except Exception as e:
                    raise DeveloperError(
                        f"After calling:"
                        f"\n\t>>> f = {Op.__name__}()"
                        f"\n\t>>> f({x}, {y})"
                        f"\n Calling:"
                        f"\n`{'f.partial_a()' if n == 0 else 'f.partial_b()'}` raised the error:"
                        f"\n\t{type(e).__name__}: {e}"
                        f"\nThis is an error with the grader, please contact an instructor")

                if not isinstance(actual, Real):
                    raise StudentError(
                        f"After calling:"
                        f"\n\t>>> f = {Op.__name__}()"
                        f"\n\t>>> f({x}, {y})"
                        f"\n`{'f.partial_a()' if n == 0 else 'f.partial_b()'}` should have returned "
                        f"the floating point number:"
                        f"\n\t{expected}"
                        f"\nInstead, it returned an object of type {type(actual)}:"
                        f"\n\t{actual}")

                if not isclose(actual, expected, rel_tol=1e-5, abs_tol=1e-5):
                    raise StudentError(
                        f"After calling:"
                        f"\n\t>>> f = {Op.__name__}()"
                        f"\n\t>>> f({x}, {y})"
                        f"\n`{'f.partial_a()' if n == 0 else 'f.partial_b()'}` should have returned:"
                        f"\n\t{expected}"
                        f"\nInstead, it returned:"
                        f"\n\t{actual}")
    print_passed(sha224(str.encode(__seed__[3])).hexdigest())



from collections.abc import Sequence
import numpy as np
from typing import Any, Optional, Callable
from math import isclose
from numbers import Real, Integral


def _same_dtype(x: np.ndarray, y: np.ndarray) -> bool:
    """
    Checks if `x` and `y` are subtypes of the same generic
    numpy data type. E.g. if `x` and `y` both store integers.

    Parameters
    ----------
    x : np.ndarray
    y : np.ndarray

    Returns
    -------
    bool
        `True` if `x` and `y` have the same generic dtype.
    """
    return np.issubdtype(x.dtype, np.floating) and np.issubdtype(y.dtype, np.floating) or \
           np.issubdtype(x.dtype, np.integer) and np.issubdtype(y.dtype, np.integer) or \
           np.issubdtype(x.dtype, np.bool_) and np.issubdtype(y.dtype, np.bool_)


def _same_abstract_numeric(x: Any, y: Any) -> bool:
    """
    Checks if `x` and `y` both numeric types and belong to the
    same abstract numeric type

    Parameters
    ----------
    x : Any
    y : Any

    Returns
    -------
    bool
    """
    if isinstance(x, bool) or isinstance(y, bool):
        return False

    if not isinstance(x, Real) or not isinstance(y, Real):
        return False

    if isinstance(x, Integral):
        return isinstance(y, Integral)

    if isinstance(y, Integral):
        return isinstance(x, Integral)
    return True


def check_arrays(actual: np.ndarray, desired: np.ndarray) -> str:
    if isinstance(actual, np.ndarray):
        if actual.shape != desired.shape:
            return f"Expected a shape-{desired.shape} array with value:\n" \
                   f"{repr(desired)}\n\n" \
                   f"Got a shape-{actual.shape} array with value:\n" \
                   f"{repr(actual)}"

        if not _same_dtype(actual, desired):
            return f"Expected an array of dtype-{desired.dtype} with value:\n" \
                   f"{repr(desired)}\n\n" \
                   f"Got an array of dtype-{actual.dtype} with value:\n" \
                   f"{repr(actual)}"

        if not np.allclose(actual, desired, atol=1e-5, rtol=1e-5):
            return f"Expected a shape-{desired.shape} array with value:\n" \
                   f"{repr(desired)}\n\n" \
                   f"Got a shape-{actual.shape} array with value:\n" \
                   f"{repr(actual)}"
    return ""


def mismatch_error(actual: Any, desired: Any,
                   comparison_function: Optional[Callable[[Any, Any], bool]] = None) -> str:
    """
    Compares two objects and returns a descriptive message
    describing the nature of their mismatched types/values.

    Returns an empty string if the two objects match.

    Parameters
    ----------
    actual : Any
        The actual object that was produced.

    desired : Any
        The desired object.

    comparison_function : Optional[Callable[[Any, Any], Union[bool, str]]]
        If specified, this will be used to compare the outputs of
        the respective functions. If it returns `False` and error
        message will print. If it returns a string, that will be printed
        as the error message.

    Returns
    -------
    mismatch_msg : str
        A message describing the nature of the mismatch between the
        actual object and the desired one. The string is empty if the
        two are found to match.

    Notes
    -----
    NumPy arrays are handled gracefully - shapes are compared up front,
    and data types are assured to be of the same parent-type (e.g.
    float64 and float32 are agreeable).
    """
    from .errors import DeveloperError

    if comparison_function is not None:
        o = comparison_function(actual, desired)
        if not isinstance(o, (bool, str)):
            raise DeveloperError(f"comparison_function({repr(actual)}, {repr(desired)}) "
                                 f"returned type `{type(o)}`.")
        else:
            return o if o else ""

    if type(actual) is not type(desired) and not _same_abstract_numeric(actual, desired):
        return f"Expected an output of type-{type(desired)} with value:\n{repr(desired)}\n\n" \
               f"Got an output of type-{type(actual)} with value:\n{repr(actual)}"

    if isinstance(actual, float) or isinstance(desired, float):
        if not isclose(actual, desired, rel_tol=1e-6, abs_tol=1e-6):
            return f"Expected:\n" \
                   f"{repr(desired)}\n\n" \
                   f"Got:\n" \
                   f"{repr(actual)}"
        else:
            return ""

    if isinstance(actual, np.ndarray):
        return check_arrays(actual, desired)

    if isinstance(desired, Sequence):
        if len(desired) != len(actual):
            return f"The expected output has length-{len(desired)}:\n" \
                   f"{repr(desired)}\n\n" \
                   f"Got an output of length-{len(actual)}:\n" \
                   f"{repr(actual)}"

    # if type(desired).__name__ not in dir(__builtins__) and not hasattr(desired, "__eq__"):
    #     return ""

    if isinstance(desired, (list, tuple)) and any(isinstance(i, np.ndarray) for i in desired):
        for n, (desired_item, actual_item) in enumerate(zip(desired, actual)):
            msg = mismatch_error(desired=desired_item, actual=actual_item)
            if msg:
                return f"Element {n} of the student's solution " \
                       f"does not match the corresponding element of the solution:\n{msg}." \
                       f"\n\nThe full expected solution is:\n"\
                       f"{repr(desired)}\n\n" \
                       f"Got:\n" \
                       f"{repr(actual)}"
    elif actual != desired:
        return f"Expected:\n" \
               f"{repr(desired)}\n\n" \
               f"Got:\n" \
               f"{repr(actual)}"
    return ""



from typing import *
from numpy import ndarray
import numpy as np


__all__ = ["grade_distances", "grade_predict", "grade_make_folds"]
__seed__ = ["distances", "predict", "makefolds"]


def grade_distances(student_func: Callable[[ndarray, ndarray], ndarray]):
    from bwsi_grader.print import print_passed
    from bwsi_grader import compare_functions
    from hashlib import sha224

    x = np.array([[1., -3., 8., 5.],  # shape-(2, 4)
                  [10., 3., 12., 0.]])

    y = np.array([[1., -3., 8., 5.],  # shape-(3, 4)
                  [9.,  0., 5., 2.],
                  [22., -1., -12., 0.]])

    def soln(x, y):
        dists = -2 * np.dot(x, y.T)
        dists += np.sum(x ** 2, axis=1, keepdims=True)
        dists += np.sum(y ** 2, axis=1)
        dists = np.sqrt(dists, out=dists)
        return dists

    # test static case
    compare_functions(student=student_func, soln=soln, fn_args=(x, y))

    for i in range(1000):
        M, N, D = np.random.randint(1, 6, size=3)
        x = 10 * np.round(np.random.rand(M, D), 2)
        y = 10 * np.round(np.random.rand(N, D), 2)
        compare_functions(student=student_func, soln=soln, fn_args=(x, y))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def grade_predict(student_func: Callable[[ndarray, ndarray, int], ndarray]):
    from bwsi_grader.print import print_passed
    from bwsi_grader import compare_functions
    from hashlib import sha224

    def soln(dists, labels, k=1):
        import numpy as np

        def _mode(seq):
            from collections import Counter
            from itertools import takewhile
            most_common = Counter(seq).most_common()
            _max = most_common[0][1]
            return min(item for item, cnt in takewhile(lambda x: x[1] == _max, most_common))

        label_sort = np.argsort(labels)
        dists = dists[:, label_sort]
        labels = labels[label_sort]

        k_nearest = np.argsort(dists, axis=1)
        labels = labels[k_nearest]

        k_vals = (np.isclose(dists, dists.min(axis=1, keepdims=True))).sum(axis=1)
        k_vals[k_vals < k] = k
        return np.asarray([_mode(row[:k_val]) for row, k_val in zip(labels, k_vals)])

    dists = np.array([[1., 0.1, 3., 2., 10., 1.5, 12.]])
    labels = np.array([2,    0,  2,  2,  3,    8,  5])
    k = 5
    compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    # tie-breaker
    dists = np.array([[0.1, 0.1]])
    labels = np.array([2,  0])
    k = 2
    compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    # tie-breaker
    dists = np.array([[0.1, 0.1, 0.1]])
    labels = np.array([3,  1, 10])
    k = 3
    compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    # tie-breaker
    dists = np.array([[0.1, 0.1, 0.1, .8, .5]])
    labels = np.array([  3,   1,  10,  2,  2])
    k = 5
    compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    dists = np.array([[0., 1., 1., 1.]])
    labels = np.array([0, 1, 1, 0])
    k = 3
    compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    for i in range(100):
        M, N = np.random.randint(1, 6, size=2)
        k = int(np.random.randint(1, N + 1))
        dists = 10 * abs(np.round(np.random.rand(M, N), 2))
        labels = np.random.randint(0, 10, size=N)
        compare_functions(student=student_func, soln=soln, fn_args=(dists, labels, k))

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())


def grade_make_folds(student_func: Callable[[ndarray, int], List[ndarray]]):
    from bwsi_grader.print import print_passed
    from bwsi_grader import compare_functions
    from hashlib import sha224

    def soln(x, num_folds):
        fold_size = x.shape[0]//num_folds
        return [x[i*fold_size:(i+1)*fold_size] for i in range(num_folds)]

    x = np.array([1, 2, 3, 4, 5, 6])

    for num_folds in range(1, 7):
        compare_functions(student=student_func, soln=soln, fn_args=(x, num_folds))

    y = np.array([[ 0,  1],
                  [ 2,  3],
                  [ 4,  5],
                  [ 6,  7],
                  [ 8,  9],
                  [10, 11]]).astype(float)

    for num_folds in range(1, 4):
        compare_functions(student=student_func, soln=soln, fn_args=(y, num_folds))

    print_passed(sha224(str.encode(__seed__[2])).hexdigest())




__seed__ = ['file_parser']


def grade_file_parser(student_fuction):
    """ Test solution to "Odds and Ends"

    Parameters
    ----------
    student_fuction: Callable[str]
        Student solution function.
    """
    from bwsi_grader.print import print_passed
    from bwsi_grader import compare_functions
    from hashlib import sha224
    import numpy as np
    import os

    def write_test_file(fname, list_length=10, items_per_entry=3):
        """ Write test file used to evaluate student function.

        Parameters
        ----------
        fname: str
            File name.

        list_length: int, optional (default=10)
            Length of file.

        items_per_entry: int, optional (default=3)
            Max number of entries (food-category pairs) per line.
        """
        foods = {'dessert': ['cake', 'cookies', 'ice cream', 'candy'],
                 'vegetable': ['carrots', 'spinach', 'broccoli', 'kale'],
                 'meat': ['chicken', 'pork', 'steak', 'tuna'],
                 'fruit': ['apples', 'peaches', 'pears', 'bananas']}

        entries = []
        for key, values in foods.items():
            for value in values:
                entries.extend([f'{value}: {key}'])

        final_entries = []
        for i in range(list_length):
            choices = np.random.choice(entries, np.random.randint(1, items_per_entry), replace=False)
            final_entries.append(', '.join(choices))

        with open(fname, 'w') as f:
            f.write('\n'.join(final_entries))

    def solution(file_path):
        """ Read in survey and determine the most common food of each type.

        Parameters
        ----------
        file_path : str
            Path to text file containing favorite food survey responses.

        Returns
        -------
        Dict[str, str]
            Dictionary with the key being food type and value being food.
        """
        from collections import Counter
        from itertools import chain

        with open(file_path, 'r') as f:
            # read correctly formatted lines
            entries = list(filter(lambda x: ': ' in x, chain(*[entry.split(', ') for entry in f.read().splitlines()])))

        responses = {}
        for entry in entries:
            food, food_type = entry.split(': ')
            responses.setdefault(food_type, []).append(food)

        return {k: Counter(sorted(v)).most_common()[0][0] for (k, v) in responses.items()}

    # generate random files and compare results
    foods_file = 'favorite-foods.txt'
    try:
        for i in range(100):
            write_test_file(foods_file, list_length=np.random.randint(1, 50), items_per_entry=np.random.randint(2, 5))
            compare_functions(student=student_fuction, soln=solution, fn_args=(foods_file,))
    finally:
        os.remove(foods_file)

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())



"""
In this problem, we will learn about a simple algorithm for removing noise from ('denoising')
an image. We will want to use vectorization to write an efficient algorithm for this.
"""

from typing import *
from numpy import ndarray

import numpy as np

__all__ = ["grader1", "grader2", "grader3"]


def print_passed(hash_str: str):
    msg = f"""
============================== ALL TESTS PASSED! ===============================
                                  Great Job!
================================================================================
"""
    print(msg)


def grader1(student_func: Callable[[ndarray], ndarray]):
    """ Grades Problem 1: local energy of each pixel """

    from bwsi_grader import compare_functions
    print("Please note that these problem are *not* graded. The autograder "
          "will tell you if your solution is correct, but will not provide a hash."
          "The solution notebook is available on EdX.")

    def soln(img):
        ''' Given a 2D array of color values, produces a 2D array with
            the energy at each pixel, where energy is defined as the sum
            of each's pixels neighbors differing in color from that pixel.

            Parameters
            ----------
            img : numpy.ndarray, shape=(M, N)
                An MxN array of color values.

            Returns
            -------
            numpy.ndarray, shape=(M, N)
                An MxN array of energy values.
        '''
        energies = np.zeros_like(img)

        energies[:-1] += img[:-1] != img[1:]           # below
        energies[1:] += img[1:] != img[:-1]            # above
        energies[:, :-1] += img[:, :-1] != img[:, 1:]  # right
        energies[:, 1:] += img[:, 1:] != img[:, :-1]   # left

        return energies

    # check standard case
    img = np.array([[0, 0, 0, 0, 1, 1, 0, 1],
                    [0, 1, 0, 0, 1, 1, 1, 1],
                    [0, 0, 0, 0, 0, 1, 1, 1],
                    [0, 0, 0, 0, 1, 1, 1, 1],
                    [0, 0, 1, 0, 1, 1, 1, 1]])
    compare_functions(student=student_func, soln=soln, fn_args=(img,))

    # check random cases
    for i in range(100):
        shape = np.random.randint(3, 7, 2)
        img = np.random.randint(0, 255, size=shape).astype('uint8')
        compare_functions(student=student_func, soln=soln, fn_args=(img,))

    print_passed("Great job!" + "!"*46)


def grader2(student_func: Callable[[ndarray, Tuple[int, int]], List[int]]):
    """ Grades Problem 2: neighbor colors."""

    from bwsi_grader import compare_functions


    def soln(img, pixel):
        ''' Given a 2D array of color values and the position of a pixel,
            returns a list of `pixel`'s neighboring color values.

        Parameters
        ----------
        img : numpy.ndarray, shape=(M, N)
            An MxN array of color values
        pixel : tuple[int, int]
            The (r, c) index of the pixel whose neighbors to retrieve.

        Returns
        -------
        List[int]
            The color (or label) value of each of `pixel`'s neighbors.
        '''
        neighbor_vals = []
        if pixel[0] > 0:
            neighbor_vals.append(img[pixel[0]-1, pixel[1]])
        if pixel[1] > 0:
            neighbor_vals.append(img[pixel[0], pixel[1]-1])
        if pixel[0] < img.shape[0]-1:
            neighbor_vals.append(img[pixel[0]+1, pixel[1]])
        if pixel[1] < img.shape[1]-1:
            neighbor_vals.append(img[pixel[0], pixel[1]+1])
        return sorted(neighbor_vals)

    def sort_student_func(*args):
        return sorted(student_func(*args))
    
    # check std case
    img = np.ones((3, 3))
    pixel = (1, 1)

    compare_functions(student=sort_student_func, soln=soln, fn_args=(img, pixel))

    # check random cases
    for i in range(100):
        rand_shape = tuple(np.random.randint(3, 7, 2))
        rand_r = np.random.randint(rand_shape[0]-1)
        rand_c = np.random.randint(rand_shape[1]-1)
        img = np.random.randint(0, 255, rand_shape)
        pixel = (rand_r, rand_c)
        compare_functions(student=sort_student_func, soln=soln, fn_args=(img, pixel))

    print_passed("Great job!" + "!"*46)


def grader3(student_func: Callable[[ndarray], ndarray]):
    """ Grades Problem 3: Iterated Conditional Modes"""

    from bwsi_grader import compare_functions

    def soln(img):
        img = img.copy()

        def compute_energy(img):
            energies = np.zeros_like(img)
            energies[:-1] += img[:-1] != img[1:]           # below
            energies[1:] += img[1:] != img[:-1]            # above
            energies[:, :-1] += img[:, :-1] != img[:, 1:]  # right
            energies[:, 1:] += img[:, 1:] != img[:, :-1]   # left
            return energies

        def get_neighbor_colors(img, pixel):
            neighbor_vals = []
            if pixel[0] > 0:
                neighbor_vals.append(img[pixel[0]-1, pixel[1]])
            if pixel[1] > 0:
                neighbor_vals.append(img[pixel[0], pixel[1]-1])
            if pixel[0] < img.shape[0]-1:
                neighbor_vals.append(img[pixel[0]+1, pixel[1]])
            if pixel[1] < img.shape[1]-1:
                neighbor_vals.append(img[pixel[0], pixel[1]+1])
            return neighbor_vals

        energies = compute_energy(img)
        highest_energy = np.divmod(np.argmax(energies), img.shape[1])
        neighbors = get_neighbor_colors(img, highest_energy)
        (neighbor_labels, neighbor_counts) = np.unique(neighbors, return_counts=True)
        best_label = neighbor_labels[np.argmax(neighbor_counts)]
        img[highest_energy] = best_label
        return img

    # check standard case
    img = np.array([[0, 0, 0, 0, 1, 1, 0, 1],
                    [0, 1, 0, 0, 1, 1, 1, 1],
                    [0, 0, 0, 0, 0, 1, 1, 1],
                    [0, 0, 0, 0, 1, 1, 1, 1],
                    [0, 0, 1, 0, 1, 1, 1, 1]])
    compare_functions(student=student_func, soln=soln, fn_args=(img,))

    # check random cases
    for i in range(100):
        shape = np.random.randint(3, 7, 2)
        img = np.random.randint(0, 255, size=shape).astype('uint8')
        compare_functions(student=student_func, soln=soln, fn_args=(img,))

    print_passed("Great job!" + "!"*46)



import versioneer
from setuptools import find_packages, setup

DISTNAME = 'bwsi_grader'
LICENSE = 'MIT'
AUTHOR = 'Ryan Soklaski'
AUTHOR_EMAIL = 'ry26099@mit.edu'
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Education',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
]

INSTALL_REQUIRES = ['numpy >= 1.10']
TESTS_REQUIRE = ['pytest >= 3.8', 'hypothesis >= 4.0']

DESCRIPTION = "A suite of auto-graders for the BWSI EdX courses."
LONG_DESCRIPTION = """
A package to be utilized locally by students to grade assignments that 
are written in Python. These graders correspond to coursework associated 
with BWSI-EdX.
"""


setup(name=DISTNAME,
      platforms=['Windows', 'Linux', 'Mac OS-X', 'Unix'],
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license=LICENSE,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      classifiers=CLASSIFIERS,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      install_requires=INSTALL_REQUIRES,
      tests_require=TESTS_REQUIRE,
      python_requires='>=3.6',
      packages=find_packages(exclude=["tests", "tests.*",
                                      "solutions", "solutions.*",
                                      "bwsi_solutions", "bwsi_solutions*"]))



from . import ciphers
from . import heat_dispersion
from . import image_restoration
from . import palindrome
from . import pizza_shop
from . import relaxation_method
from . import run_length_encoding
from . import three_five
from . import card_games

__all__ = ["ciphers",
           "heat_dispersion",
           "image_restoration",
           "palindrome",
           "pizza_shop",
           "relaxation_method",
           "run_length_encoding",
           "three_five",
           "card_games",
           "odds_and_ends",
           "card_games",
           ]



"""
# Run-Length Encoding
Run-length encoding is a simple method for compressing data that contains long sequences of repeated characters.

In this compression algorithm:
1. A standalone character will be unchanged. E.g `"a"` $\rightarrow$ `["a"]`.
2. A run of a character, `c`, repeated `N` times will be compressed to `["c", "c", N]`. E.g. `"bbbb"` $\rightarrow$ `['b', 'b', 4]`.

These two rules are all that you need to perform run-length encoding.

Let's look at a few examples of run-length-encoding:

- `"abcd"` $\rightarrow$ `['a', 'b', 'c', 'd']`
- `"abbbba"` $\rightarrow$ `['a', 'b', 'b', 4, 'a']`
- `"aaaabbcccd"` $\rightarrow$ `['a', 'a', 4, 'b', 'b', 2, 'c', 'c', 3, 'd']`
- `""` $\rightarrow$ `[]`
- `"1"` $\rightarrow$ `["1"]`

The decompression algorithm, run-length decoding, simply reverses this process:

- `['q', 'a', 'a', 4, 'b', 'b', 2, 'c', 'c', 3, 'd']` $\rightarrow$ `'qaaaabbcccd'`

Here, you will implement a run-length encoding and decoding algorithms. As indicated above, the run-length encoding algorithm should be able to accept a string and return a list with the appropriate string/integer entries, according to the encoding. The decoding algorithm need be able to accept a list with an encoded sequence, and return the decoded string.

You should be able to test both of your algorithms by feeding them into one another:
```python
>>> decoder(encoder("Wooooow!!!!! I'm totally getting compressed"))
"Wooooow!!!!! I'm totally getting compressed"
```

## Required Concepts
Sequence Types
- strings
- lists
- indexing & slicing

Conditional Statements
- basic if/else conditional blocks

For-Loops
- basic for-loops

Generators & Comprehensions
- the `range` generator

Basics of Functions
- the return statement
- basic function arguments


## Examples
```python
>>> encoder("abbbba")
['a', 'b', 'b', 4, 'a']

>>> decoder(['a', 'b', 'b', 4, 'a'])
'abbbba'
```

"""

from typing import Callable, Union, Iterable, List


__seed__ = ["runlengthencoding", "runlengthdecoding"]


def _gen_run(num_runs: int = 16) -> str:
    """generate random runs of characters"""
    import string
    from random import choices, randint
    chars = string.ascii_letters + string.digits + " "
    return "".join(i*randint(0, 4) for i in choices(chars, k=num_runs))


def _gen_encoded_seq() -> List[str]:
    """ randomly generates a encoded sequence"""
    import random
    import string

    def get_char(prev):
        if prev is not None:
            return ''.join(set(string.ascii_letters) - set(prev))
        else:
            return string.ascii_letters

    def single(prev): return [random.choice(get_char(prev))]

    def multi(prev):
        letter = random.choice(get_char(prev))
        number = random.randint(2, 5)
        return [letter, letter, number]

    seq = []
    prev = None
    for i in range(random.randint(0, 5)):
        item = random.choice([single, multi])(prev)
        seq.extend(item)
        prev = item[0]
    return seq


def encoder_grader(student_func: Callable[[str], List[Union[str, int]]]):
    """ Tests the student's run-length-encoding compressor.

    Parameters
    ----------
    student_func: Callable[[str], List[Union[str, int]]]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def compression_soln(in_string: str) -> List[Union[str, int]]:
        """
        Examples
        --------
        >>> compression_soln("aaabbacccc")
        ['a', 'a', 3, 'b', 'b', 2, 'a', 'c', 4]
        """
        from itertools import groupby
        out = []
        for item, group in groupby(in_string):
            cnt = sum(1 for x in group)
            if cnt == 1:
                out.append(item)
            else:
                out.extend((item, item, cnt))
        return out

    for input_ in ("abcd", "ababa", "aabbaba", "abbbbaa", "aaaabbcccd", "", "2", "Hello Worldddd!!!"):
        compare_functions(student=student_func, soln=compression_soln, fn_args=(input_,))

    for i in range(1000):
        compare_functions(student=student_func, soln=compression_soln, fn_args=(_gen_run(),))

    print_passed(sha224(str.encode(__seed__[0])).hexdigest())


def decoder_grader(student_func: Callable[[List[Union[str, int]]], str]):
    """ Tests the student's run-length-decoding compressor.

    Parameters
    ----------
    student_func: Callable[[List[Union[str, int]]], str]
       The student's solution function.
    """
    from bwsi_grader import compare_functions
    from bwsi_grader.print import print_passed
    from hashlib import sha224

    def decompression_soln(in_list: List[Union[str, int]]) -> str:
        """
        >>> decompression_soln(['a', 'a', 5, 'b', 'b', 2, 'c', 'b', 'c'])
        "aaaaabbcbc"
        """
        out = ""
        for n, item in enumerate(in_list):
            if isinstance(item, int):
                out += in_list[n - 1] * (item - 2)
            else:
                out += item
        return out

    # check some standard cases
    std_cases = [['a', 'b', 'c', 'd'],
                 ['a', 'b', 'a', 'b', 'a'],
                 ['a', 'a', 2, 'b', 'b', 2, 'a', 'b', 'a'],
                 ['a', 'b', 'b', 4, 'a', 'a', 2],
                 ['a', 'a', 4, 'b', 'b', 2, 'c', 'c', 3, 'd'],
                 [],
                 ['2'],
                 ['H', 'e', 'l', 'l', 2, 'o', ' ', 'W', 'o',
                  'r', 'l', 'd', 'd', 4, '!', '!', 3]]

    for input_ in std_cases:
        compare_functions(student=student_func, soln=decompression_soln, fn_args=(input_,))

    for i in range(1000):
        compare_functions(student=student_func, soln=decompression_soln, fn_args=(_gen_encoded_seq(),))

    print_passed(sha224(str.encode(__seed__[1])).hexdigest())
